@@ -1,8 +1,8 @@
-{-# OPTIONS -fglasgow-exts #-}
+{-# OPTIONS -fglasgow-exts -w #-}
+module Properties where
 
-import StackSet hiding (filter)
-import qualified StackSet as S (filter)
-import Operations (tile)
+import XMonad.StackSet hiding (filter)
+import qualified XMonad.StackSet as S (filter)
 
 import Debug.Trace
 import Data.Word
@@ -11,8 +11,10 @@ import Data.Ratio
 import Data.Maybe
 import System.Environment
 import Control.Exception    (assert)
+import qualified Control.Exception as C
 import Control.Monad
 import Test.QuickCheck hiding (promote)
+import System.IO.Unsafe
 import System.IO
 import System.Random hiding (next)
 import Text.Printf
@@ -28,18 +30,19 @@ import qualified Data.Map as M
 -- Some general hints for creating StackSet properties:
 --
 -- *  ops that mutate the StackSet are usually local
--- *  most ops on StackSet should either be trivially reversible, or 
+-- *  most ops on StackSet should either be trivially reversible, or
 --    idempotent, or both.
 
 --
 -- The all important Arbitrary instance for StackSet.
 --
-instance (Integral i, Integral s, Eq a, Arbitrary a, Arbitrary sd)
-        => Arbitrary (StackSet i a s sd) where
+instance (Integral i, Integral s, Eq a, Arbitrary a, Arbitrary l, Arbitrary sd)
+        => Arbitrary (StackSet i l a s sd) where
     arbitrary = do
         sz <- choose (1,10)     -- number of workspaces
         n  <- choose (0,sz-1)   -- pick one to be in focus
         sc  <- choose (1,sz)     -- a number of physical screens
+        lay <- arbitrary           -- pick any layout
         sds <- replicateM sc arbitrary
         ls <- vector sz         -- a vector of sz workspaces
 
@@ -48,8 +51,7 @@ instance (Integral i, Integral s, Eq a, Arbitrary a, Arbitrary sd)
                             else liftM Just (choose ((-1),length s-1))
                        | s <- ls ]
 
-        return $ fromList (fromIntegral n, sds,fs,ls)
-    coarbitrary = error "no coarbitrary for StackSet"
+        return $ fromList (fromIntegral n, sds,fs,ls,lay)
 
 
 -- | fromList. Build a new StackSet from a list of list of elements,
@@ -62,14 +64,14 @@ instance (Integral i, Integral s, Eq a, Arbitrary a, Arbitrary sd)
 -- 'fs' random focused window on each workspace
 -- 'xs' list of list of windows
 --
-fromList :: (Integral i, Integral s, Eq a) => (i, [sd], [Maybe Int], [[a]]) -> StackSet i a s sd
-fromList (_,_,_,[]) = error "Cannot build a StackSet from an empty list"
+fromList :: (Integral i, Integral s, Eq a) => (i, [sd], [Maybe Int], [[a]], l) -> StackSet i l a s sd
+fromList (_,_,_,[],_) = error "Cannot build a StackSet from an empty list"
 
-fromList (o,m,fs,xs) =
+fromList (o,m,fs,xs,l) =
     let s = view o $
                 foldr (\(i,ys) s ->
                     foldr insertUp (view i s) ys)
-                        (new [0..genericLength xs-1] m) (zip [0..] xs)
+                        (new l [0..genericLength xs-1] m) (zip [0..] xs)
     in foldr (\f t -> case f of
                             Nothing -> t
                             Just i  -> foldr (const focusUp) t [0..i] ) s fs
@@ -79,7 +81,7 @@ fromList (o,m,fs,xs) =
 --
 -- Just generate StackSets with Char elements.
 --
-type T = StackSet (NonNegative Int) Char Int Int
+type T = StackSet (NonNegative Int) Int Char Int Int
 
 -- Useful operation, the non-local workspaces
 hidden_spaces x = map workspace (visible x) ++ hidden x
@@ -87,7 +89,7 @@ hidden_spaces x = map workspace (visible x) ++ hidden x
 -- Basic data invariants of the StackSet
 --
 -- With the new zipper-based StackSet, tracking focus is no longer an
--- issue: the data structure enforces focus by construction. 
+-- issue: the data structure enforces focus by construction.
 --
 -- But we still need to ensure there are no duplicates, and master/and
 -- the xinerama mapping aren't checked by the data structure at all.
@@ -129,9 +131,9 @@ monotonic (x:y:zs) | x == y-1  = monotonic (y:zs)
 prop_invariant = invariant
 
 -- and check other ops preserve invariants
-prop_empty_I  (n :: Positive Int) = forAll (choose (1,fromIntegral n)) $  \m ->
-                                    forAll (vector m) $ \ms ->
-        invariant $ new [0..fromIntegral n-1] ms
+prop_empty_I  (n :: Positive Int) l = forAll (choose (1,fromIntegral n)) $  \m ->
+                                      forAll (vector m) $ \ms ->
+        invariant $ new l [0..fromIntegral n-1] ms
 
 prop_view_I (n :: NonNegative Int) (x :: T) =
     n `tagMember` x ==> invariant $ view (fromIntegral n) x
@@ -141,6 +143,8 @@ prop_greedyView_I (n :: NonNegative Int) (x :: T) =
 
 prop_focusUp_I (n :: NonNegative Int) (x :: T) =
     invariant $ foldr (const focusUp) x [1..n]
+prop_focusMaster_I (n :: NonNegative Int) (x :: T) =
+    invariant $ foldr (const focusMaster) x [1..n]
 prop_focusDown_I (n :: NonNegative Int) (x :: T) =
     invariant $ foldr (const focusDown) x [1..n]
 
@@ -167,21 +171,24 @@ prop_swap_right_I (n :: NonNegative Int) (x :: T) =
 prop_shift_I (n :: NonNegative Int) (x :: T) =
     n `tagMember` x ==> invariant $ shift (fromIntegral n) x
 
+prop_shift_win_I (n :: NonNegative Int) (w :: Char) (x :: T) =
+    n `tagMember` x && w `member` x ==> invariant $ shiftWin (fromIntegral n) w x
+
 
 -- ---------------------------------------------------------------------
 -- 'new'
 
 -- empty StackSets have no windows in them
-prop_empty (EmptyStackSet x) = 
+prop_empty (EmptyStackSet x) =
         all (== Nothing) [ stack w | w <- workspace (current x)
                                         : map workspace (visible x) ++ hidden x ]
 
 -- empty StackSets always have focus on first workspace
-prop_empty_current (NonEmptyNubList ns) (NonEmptyNubList sds) =
+prop_empty_current (NonEmptyNubList ns) (NonEmptyNubList sds) l =
     -- TODO, this is ugly
     length sds <= length ns ==>
     tag (workspace $ current x) == head ns
-    where x = new ns sds :: T
+    where x = new l ns sds :: T
 
 -- no windows will be a member of an empty workspace
 prop_member_empty i (EmptyStackSet x)
@@ -196,7 +203,7 @@ prop_view_current (x :: T) (n :: NonNegative Int) = i `tagMember` x ==>
   where
     i = fromIntegral n
 
--- view *only* sets the current workspace, and touches Xinerama. 
+-- view *only* sets the current workspace, and touches Xinerama.
 -- no workspace contents will be changed.
 prop_view_local  (x :: T) (n :: NonNegative Int) = i `tagMember` x ==>
     workspaces x == workspaces (view i x)
@@ -229,7 +236,7 @@ prop_greedyView_current (x :: T) (n :: NonNegative Int) = i `tagMember` x ==>
   where
     i = fromIntegral n
 
--- greedyView *only* sets the current workspace, and touches Xinerama. 
+-- greedyView *only* sets the current workspace, and touches Xinerama.
 -- no workspace contents will be changed.
 prop_greedyView_local  (x :: T) (n :: NonNegative Int) = i `tagMember` x ==>
     workspaces x == workspaces (greedyView i x)
@@ -287,13 +294,16 @@ prop_index_length (x :: T) =
 --
 
 -- master/focus
--- 
+--
 -- The tiling order, and master window, of a stack is unaffected by focus changes.
 --
 prop_focus_left_master (n :: NonNegative Int) (x::T) =
     index (foldr (const focusUp) x [1..n]) == index x
 prop_focus_right_master (n :: NonNegative Int) (x::T) =
     index (foldr (const focusDown) x [1..n]) == index x
+prop_focus_master_master (n :: NonNegative Int) (x::T) =
+    index (foldr (const focusMaster) x [1..n]) == index x
+
 prop_focusWindow_master (n :: NonNegative Int) (x :: T) =
     case peek x of
         Nothing -> True
@@ -305,6 +315,9 @@ prop_focusWindow_master (n :: NonNegative Int) (x :: T) =
 prop_focus_left  (x :: T) = (focusUp  (focusDown x)) == x
 prop_focus_right (x :: T) = (focusDown (focusUp  x)) ==  x
 
+-- focus master is idempotent
+prop_focusMaster_idem (x :: T) = focusMaster x == focusMaster (focusMaster x)
+
 -- focusWindow actually leaves the window focused...
 prop_focusWindow_works (n :: NonNegative Int) (x :: T) =
     case peek x of
@@ -323,7 +336,10 @@ prop_focus_all_r (x :: T) = (foldr (const focusDown) x [1..n]) == x
 --     f x' = foldr (\_ y -> rotate GT y) x' [1..n]
 
 -- focus is local to the current workspace
-prop_focus_local (x :: T) = hidden_spaces (focusDown x) == hidden_spaces x
+prop_focus_down_local (x :: T) = hidden_spaces (focusDown x) == hidden_spaces x
+prop_focus_up_local (x :: T) = hidden_spaces (focusUp x) == hidden_spaces x
+
+prop_focus_master_local (x :: T) = hidden_spaces (focusMaster x) == hidden_spaces x
 
 prop_focusWindow_local (n :: NonNegative Int) (x::T ) =
     case peek x of
@@ -333,19 +349,21 @@ prop_focusWindow_local (n :: NonNegative Int) (x::T ) =
                    in hidden_spaces (focusWindow (s !! i) x) == hidden_spaces x
 
 -- ---------------------------------------------------------------------
--- member/findIndex
+-- member/findTag
 
 --
--- For all windows in the stackSet, findIndex should identify the
+-- For all windows in the stackSet, findTag should identify the
 -- correct workspace
 --
 prop_findIndex (x :: T) =
-    and [ tag w == fromJust (findIndex i x)
+    and [ tag w == fromJust (findTag i x)
         | w <- workspace (current x) : map workspace (visible x)  ++ hidden x
         , t <- maybeToList (stack w)
         , i <- focus t : up t ++ down t
         ]
 
+prop_allWindowsMember w (x :: T) = (w `elem` allWindows x) ==> member w x
+
 -- ---------------------------------------------------------------------
 -- 'insert'
 
@@ -438,7 +456,7 @@ prop_delete_focus_not_end (x :: T) =
 -- preserve order
 prop_filter_order (x :: T) =
     case stack $ workspace $ current x of
-    	Nothing -> True
+        Nothing -> True
         Just s@(Stack i _ _) -> integrate' (S.filter (/= i) s) == filter (/= i) (integrate' (Just s))
 
 -- ---------------------------------------------------------------------
@@ -493,6 +511,113 @@ prop_shift_reversible i (x :: T) =
         y = swapMaster x
         n = tag (workspace $ current y)
 
+-- ---------------------------------------------------------------------
+-- shiftWin
+
+-- shiftWin on current window is the same as shift
+prop_shift_win_focus i (x :: T) =
+    i `tagMember` x ==> case peek x of
+                          Nothing -> True
+                          Just w  -> shiftWin i w x == shift i x
+
+-- shiftWin on a non-existant window is identity
+prop_shift_win_indentity i w (x :: T) =
+    i `tagMember` x && not (w  `member` x) ==> shiftWin i w x == x
+
+-- shiftWin leaves the current screen as it is, if neither i is the tag
+-- of the current workspace nor w on the current workspace
+prop_shift_win_fix_current i w (x :: T) =
+    i `tagMember` x && w `member` x && i /= n && findTag w x /= Just n
+        ==> (current $ x) == (current $ shiftWin i w x)
+    where
+        n = tag (workspace $ current x)
+
+------------------------------------------------------------------------
+-- properties for the floating layer:
+
+prop_float_reversible n (x :: T) =
+    n `member` x ==> sink n (float n geom x) == x
+        where
+            geom = RationalRect 100 100 100 100
+
+-- check rectanges were set
+{-
+prop_float_sets_geometry n (x :: T) =
+    n `member` x ==> let y = float n geom x in M.lookup y (floating x) == Just geom
+        where
+            geom = RationalRect 100 100 100 100
+-}
+
+------------------------------------------------------------------------
+
+prop_screens (x :: T) = n `elem` screens x
+ where
+    n = current x
+
+prop_differentiate xs =
+        if null xs then differentiate xs == Nothing
+                   else (differentiate xs) == Just (Stack (head xs) [] (tail xs))
+    where _ = xs :: [Int]
+
+-- looking up the tag of the current workspace should always produce a tag.
+prop_lookup_current (x :: T) = lookupWorkspace scr x == Just tg
+    where
+        (Screen (Workspace tg  _ _) scr _) = current x
+
+-- looking at a visible tag
+prop_lookup_visible (x :: T) =
+        visible x /= [] ==>
+        fromJust (lookupWorkspace scr x) `elem` tags
+    where
+        tags = [ tag (workspace y) | y <- visible x ]
+        scr = last [ screen y | y <- visible x ]
+
+
+-- ---------------------------------------------------------------------
+-- testing for failure
+
+-- and help out hpc
+prop_abort x = unsafePerformIO $ C.catch (abort "fail")
+                                         (\e -> return $  show e == "xmonad: StackSet: fail" )
+   where
+     _ = x :: Int
+
+-- new should fail with an abort
+prop_new_abort x = unsafePerformIO $ C.catch f
+                                         (\e -> return $ show e == "xmonad: StackSet: non-positive argument to StackSet.new" )
+   where
+     f = new undefined{-layout-} [] [] `seq` return False
+
+     _ = x :: Int
+
+-- prop_view_should_fail = view {- with some bogus data -}
+
+-- screens makes sense
+prop_screens_works (x :: T) = screens x == current x : visible x
+
+------------------------------------------------------------------------
+-- renaming tags
+
+-- | Rename a given tag if present in the StackSet.
+-- 408 renameTag :: Eq i => i -> i -> StackSet i l a s sd -> StackSet i l a s sd
+
+prop_rename1 (x::T) o n = o `tagMember` x && not (n `tagMember` x) ==>
+    let y = renameTag o n x
+            in n `tagMember` y
+
+prop_ensure (x :: T) l xs = let y = ensureTags l xs x
+                                in and [ n `tagMember` y | n <- xs ]
+
+prop_mapWorkspaceId (x::T) = x == mapWorkspace id x
+
+prop_mapWorkspaceInverse (x::T) = x == mapWorkspace predTag (mapWorkspace succTag x)
+  where predTag w = w { tag = pred $ tag w }
+        succTag w = w { tag = succ $ tag w }
+
+prop_mapLayoutId (x::T) = x == mapLayout id x
+
+prop_mapLayoutInverse (x::T) = x == mapLayout pred (mapLayout succ x)
+
 ------------------------------------------------------------------------
 -- some properties for layouts:
 
@@ -500,7 +625,7 @@ prop_shift_reversible i (x :: T) =
 {-
 prop_tile_fullscreen rect = tile pct rect 1 1 == [rect]
 
--- multiple windows 
+-- multiple windows
 prop_tile_non_overlap rect windows nmaster = noOverlaps (tile pct rect nmaster windows)
   where _ = rect :: Rectangle
 
@@ -526,7 +651,7 @@ noOverlaps xs  = and [ verts a `notOverlap` verts b
 
 main :: IO ()
 main = do
-    args <- getArgs
+    args <- fmap (drop 1) getArgs
     let n = if null args then 100 else read (head args)
     (results, passed) <- liftM unzip $ mapM (\(s,a) -> printf "%-25s: " s >> a n) tests
     printf "Passed %d tests!\n" (sum passed)
@@ -561,20 +686,27 @@ main = do
         ,("index/length"        , mytest prop_index_length)
 
         ,("focus left : invariant", mytest prop_focusUp_I)
+        ,("focus master : invariant", mytest prop_focusMaster_I)
         ,("focus right: invariant", mytest prop_focusDown_I)
         ,("focusWindow: invariant", mytest prop_focus_I)
         ,("focus left/master"   , mytest prop_focus_left_master)
         ,("focus right/master"  , mytest prop_focus_right_master)
+        ,("focus master/master"  , mytest prop_focus_master_master)
         ,("focusWindow master"  , mytest prop_focusWindow_master)
         ,("focus left/right"    , mytest prop_focus_left)
         ,("focus right/left"    , mytest prop_focus_right)
         ,("focus all left  "    , mytest prop_focus_all_l)
         ,("focus all right "    , mytest prop_focus_all_r)
-        ,("focus is local"      , mytest prop_focus_local)
+        ,("focus down is local"      , mytest prop_focus_down_local)
+        ,("focus up is local"      , mytest prop_focus_up_local)
+        ,("focus master is local"      , mytest prop_focus_master_local)
+        ,("focus master idemp"  , mytest prop_focusMaster_idem)
+
         ,("focusWindow is local", mytest prop_focusWindow_local)
         ,("focusWindow works"   , mytest prop_focusWindow_works)
 
-        ,("findIndex"           , mytest prop_findIndex)
+        ,("findTag"           , mytest prop_findIndex)
+        ,("allWindows/member"   , mytest prop_allWindowsMember)
 
         ,("insert: invariant"   , mytest prop_insertUp_I)
         ,("insert/new"          , mytest prop_insert_empty)
@@ -611,6 +743,30 @@ main = do
 
         ,("shift: invariant"    , mytest prop_shift_I)
         ,("shift is reversible" , mytest prop_shift_reversible)
+        ,("shiftWin: invariant" , mytest prop_shift_win_I)
+        ,("shiftWin is shift on focus" , mytest prop_shift_win_focus)
+        ,("shiftWin fix current" , mytest prop_shift_win_fix_current)
+
+        ,("floating is reversible" , mytest prop_float_reversible)
+        ,("screens includes current", mytest prop_screens)
+        ,("differentiate works", mytest prop_differentiate)
+        ,("lookupTagOnScreen", mytest prop_lookup_current)
+        ,("lookupTagOnVisbleScreen", mytest prop_lookup_visible)
+        ,("screens works",      mytest prop_screens_works)
+        ,("renaming works",     mytest prop_rename1)
+        ,("ensure works",     mytest prop_ensure)
+
+        ,("mapWorkspace id", mytest prop_mapWorkspaceId)
+        ,("mapWorkspace inverse", mytest prop_mapWorkspaceInverse)
+        ,("mapLayout id", mytest prop_mapLayoutId)
+        ,("mapLayout inverse", mytest prop_mapLayoutInverse)
+
+        -- testing for failure:
+        ,("abort fails",            mytest prop_abort)
+        ,("new fails with abort",   mytest prop_new_abort)
+        ,("shiftWin identity",      mytest prop_shift_win_indentity)
+
+        -- renaming
 
 {-
         ,("tile 1 window fullsize", mytest prop_tile_fullscreen)
@@ -781,8 +937,10 @@ instance Arbitrary EmptyStackSet where
     arbitrary = do
         (NonEmptyNubList ns)  <- arbitrary
         (NonEmptyNubList sds) <- arbitrary
+        l <- arbitrary
         -- there cannot be more screens than workspaces:
-        return . EmptyStackSet . new ns $ take (min (length ns) (length sds)) sds
+        return . EmptyStackSet . new l ns $ take (min (length ns) (length sds)) sds
+    coarbitrary = error "coarbitrary EmptyStackSet"
 
 -- | Generates a value that satisfies a predicate.
 suchThat :: Gen a -> (a -> Bool) -> Gen a
